/*
** Module   : WidgetConfig.java
** Abstract : Base class of a widget configuration.
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140924 Created initial version.
** 002 HC  20141021 Implemented window move/position support.
** 003 VIG 20150128 Added popupMenuId field.
** 004 CA  20150228 If the enable state has changed, update the widget too.
** 005 CA  20150323 Added field to store the widget's value set at the frame definition. Useful 
**                  for literal widgets.
** 006 HC  20150323 Removed business logic from config classes.
** 007 VIG 20150407 The attribute "modified" was moved from BaseConfig.
*/ 

package com.goldencode.p2j.ui;

import java.io.*;

import org.aspectj.lang.*;

import com.goldencode.p2j.aspects.ui.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

/**
 * Base class for all widget configuration classes.  All sub-classes which require instrumentation
 * by the {@link ConfigFieldSetterAspect} class must have as suffix the <code>Config</code> text.
 * See the javadoc for {@link ConfigFieldSetterAspect#beforeSetField} for details about which
 * fields are instrumented.
 * <p>
 * All widget state needs to be separated from the concrete widget implementation (be it on server
 * or client-side).  This ensures the state is shared between devices (i.e. a ChUI and GUI impl of
 * a widget can access the state) and on server-side ensures that shared widgets (TODO: P2J 
 * doesn't create unique widgets resources for a child shared frame) can share this state, too.
 * <p>
 * Explicit setters and getters can be defined only as a last resort.  Before deciding to add a
 * setter/getter here, determine if the case is related to server-side or client-side, and check
 * if the setter/getter at the widget implementation can address this problem, instead of the
 * widget config class.
 * <p>
 * Only public, non-transient, instance fields will be collected by the client-side and pushed to
 * the server-side.  The server-side will always send the entire definition (via pushWidgetConfig
 * calls).
 * <p>
 * This class implements {@link Externalizable}, to allow the server-side to send it to the 
 * client-side in one go.  All sub-classes need to ensure that the {@link #writeExternal} and 
 * {@link #readExternal} are overridden, in case the config class has additional instance fields.
 * <p>
 * All sub-classes of this class must be added to the {@link ClientConfigManager#CONFIG_CLASSES} 
 * list.
 * <p>
 * The following are some TODO's to improve the usage of the {@link WidgetConfig} classes:
 * <ul>
 * <li>Implement a listener/notification framework and a clean way to associate listeners with 
 *     their target changes. This would simplify things like the server-side 
 *     {@link LogicalTerminal#pushScreenDefinition} or the scaling that we do via 
 *     {@link Coordinate#scale}.  WARNING: if AspectJ is used, be careful that it can affect
 *     performance, if the called aspect has any parameters (like {@link JoinPoint}) or if the
 *     called aspect is time-consuming.  In any case, once this is added, the performance needs
 *     to be checked if it is affected by the changes.</li>
 *  <li>
 * <li>On client-side, eliminate the {@link AbstractWidget#config} method and any reliance on 
 *     direct access to widget configs from outside of widgets. Move all widget-internal access 
 *     to direct field references on the local properly-typed config instance.</li>
 * <li>Move all client-state in the associated {@link WidgetConfig} instances, to allow complete 
 *     separation between a widget's state and its concrete implementation (as a widget may exist
 *     for GUI or ChUI at the same time). WARNING: this client-side specific state must not be
 *     instrumented by the {@link ConfigFieldSetterAspect} and it must not be sent to server-side.
 * </li>
 * </ul>
 */
public abstract class WidgetConfig
implements Externalizable
{
   /** Serial Version ID */
   private static final long serialVersionUID = 6073071302160174183L;

   /** The index of the frame which holds this widget. */
   public int frameId = -1;
   
   /** Means whether the widget is visible or not. */
   public boolean visible = false; 
   
   /** Flag to indicate whether the widget has been realized. **/
   public boolean realized = false;

   /** The HIDDEN attribute, which stops widgets from displaying. */
   public boolean hidden = false; 
   
   /** Mark header widget. */
   public boolean header = false;

   /** Widget enabled flag. */
   public boolean enabled = true;
   
   /** An ID of POPUP-MENU assigned to widget or -1 */
   public int popupMenuId = -1; 

   /** MODIFIED attribute variable. */
   public boolean modified = false;

   /** Widget ID. */
   public WidgetId id = null;

   /**
    * The state of the data for this widget.  Do not set directly on client-side.  Instead, use
    * the {@link AbstractWidget#setState} API.
    */ 
   public byte state = ScreenBuffer.UNINITIALIZED;

   /** Store the widget's value set at frame definition. */
   public BaseDataType frameDefValue = null;
   
   /**
    * Default constructor (only used in deserialization).
    */
   public WidgetConfig()
   {
   }
   
   /**
    * Create a new config and associate it with the given widget.
    * 
    * @param    id
    *           The ID of the widget to which this configuration belongs.
    */
   protected WidgetConfig(int id)
   {
      this(new WidgetId(id));
   }
   
   /**
    * Create a new config and associate it with the given widget.
    * 
    * @param    id
    *           The ID of the widget to which this configuration belongs.
    */
   protected WidgetConfig(WidgetId id)
   {
      this.id = id;
   }
   
   /**
    * Set the <code>header</code> attribute in this widget configuration, dynamically.
    * <p>
    * If the configuration doesn't define a field with the <code>header</code> name, then this
    * will be a no-op.
    * 
    * @param    header
    *           The new value for the header attribute.
    */
   public void setDynamicHeader(boolean header)
   {
      ConfigManager.getInstance().setDynamicConfig(this, "header", boolean.class, header);
   }

   /**
    * Set the <code>LABELS</code> attribute in this widget configuration, dynamically.
    * <p>
    * If the configuration doesn't define a field with the <code>labels</code> name, then this
    * will be a no-op.
    * 
    * @param    labels
    *           The new value for the LABELS attribute.
    */
   public void setDynamicLabels(boolean labels)
   {
      ConfigManager.getInstance().setDynamicConfig(this, "labels", boolean.class, labels);
   }
   
   /**
    * Get the <code>FORMAT</code> attribute in this widget configuration, dynamically.
    * <p>
    * If the configuration doesn't define a field with the <code>format</code> name, then this
    * will return <code>null</code>
    * 
    * @return   See above.
    */
   public String getDynamicFormat()
   {
      return ConfigManager.getInstance().getDynamicConfig(this, "format", String.class, null);
   }
   
   /**
    * Set the <code>FORMAT</code> attribute in this widget configuration, dynamically.
    * <p>
    * If the configuration doesn't define a field with the <code>format</code> name, then this
    * will be a no-op.
    * 
    * @param    format
    *           The new value for the FORMAT attribute.
    */
   public void setDynamicFormat(String format)
   {
      ConfigManager.getInstance().setDynamicConfig(this, "format", String.class, format);
   }

   /**
    * Get the <code>COLUMN-LABEL</code> attribute in this widget configuration, dynamically.
    * <p>
    * If the configuration doesn't define a field with the <code>clabel</code> name, then this
    * will return <code>null</code>
    * 
    * @return   See above.
    */
   public String getDynamicColumnLabel()
   {
      return ConfigManager.getInstance().getDynamicConfig(this, "clabel", String.class, null);
   }
   
   /**
    * Get the <code>LABEL</code> attribute in this widget configuration, dynamically.
    * <p>
    * If the configuration doesn't define a field with the <code>label</code> name, then this
    * will return <code>null</code>
    * 
    * @return   See above.
    */
   public String getDynamicLabel()
   {
      return ConfigManager.getInstance().getDynamicConfig(this, "label", String.class, null);
   }
   
   /**
    * Set the <code>LABEL</code> attribute in this widget configuration, dynamically.
    * <p>
    * If the configuration doesn't define a field with the <code>label</code> name, then this
    * will be a no-op.
    * 
    * @param    label
    *           The new value for the LABEL attribute.
    */
   public void setDynamicLabel(String label)
   {
      ConfigManager.getInstance().setDynamicConfig(this, "label", String.class, label);
   }
   

   /**
    * Get the <code>HELP</code> attribute in this widget configuration, dynamically.
    * <p>
    * If the configuration doesn't define a field with the <code>help</code> name, then this
    * will return <code>null</code>
    * 
    * @return   See above.
    */
   public String getDynamicHelp()
   {
      return ConfigManager.getInstance().getDynamicConfig(this, "help", String.class, null);
   }
   
   /**
    * Set the <code>HELP</code> attribute in this widget configuration, dynamically.
    * <p>
    * If the configuration doesn't define a field with the <code>help</code> name, then this
    * will be a no-op.
    * 
    * @param    help
    *           The new value for the HELP attribute.
    */
   public void setDynamicHelp(String help)
   {
      ConfigManager.getInstance().setDynamicConfig(this, "help", String.class, help);
   }
   
   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   @Override
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      out.writeObject(id);
      out.writeInt(frameId);
      out.writeInt(popupMenuId);
      out.writeBoolean(enabled);
      out.writeBoolean(visible);
      out.writeBoolean(realized);
      out.writeBoolean(header);
      out.writeBoolean(hidden);
      out.writeBoolean(modified);
      out.writeObject(frameDefValue);
   }

   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   @Override
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      id          = (WidgetId) in.readObject();
      frameId     = in.readInt();
      popupMenuId = in.readInt();
      enabled     = in.readBoolean();
      visible     = in.readBoolean();
      realized    = in.readBoolean();
      header      = in.readBoolean();
      hidden      = in.readBoolean();
      modified    = in.readBoolean();
      frameDefValue = (BaseDataType) in.readObject();
   }
   
   /**
    * Get the string representation of this widget configuration.
    * 
    * @return   See above.
    */
   @Override
   public String toString()
   {
      // on server-side, get only the ID and the class
      return (id == null ? "<null>" : id.toString()) + ": " + this.getClass();
   }

   /**
    * Set new values from the provided configuration instance.
    * 
    * @param    config
    *           The instance from which to take values.
    */
   public void applyConfig(WidgetConfig config)
   {
      // do not copy the ID.
      
      frameId = config.frameId;
      popupMenuId = config.popupMenuId;
      visible = config.visible;
      realized = config.realized;
      hidden = config.hidden;
      header = config.header;
      enabled = config.enabled;
      state = config.state;
      ConfigHelper.setModified(this, config.modified);
      frameDefValue = config.frameDefValue;
   }
}
