/*
** Module   : WindowManager.java
** Abstract :
**
** Copyright (c) 2010-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SIY 20100629 Created initial version
** 002 SIY 20100804 Changes in imports, some refactorings.
** 003 GES 20111012 Moved static data to context-local. This allows multiple client sessions to be 
**                  hosted in the server.
** 004 SIY 20111202 Updated imports.
** 005 CA  20140805 Changes related to multi-window support and GUI windows.
** 006 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 007 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 008 CA  20141123 First pass at window clearing during client restart in GUI.  Added support 
**                  for default location and dimension, for dynamic windows.
** 009 CA  20141125 When the window is destroyed, this needs to be done outside of the synchronized
**                  block. 
** 010 CA  20150408 Added API to compute the children for a window.
**                  Keep the window's original order, so that the window tree can be identified 
**                  correctly.
** 011 HC  20150323 Removed business logic from config classes.
** 012 VIG 20150401 Fixed null child focus.
*/

package com.goldencode.p2j.ui.client;

import java.util.*;

import com.goldencode.p2j.security.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Window manager.
 */
public class WindowManager
{
   /** Stores context-local state variables. */
   private static final ContextLocal<WorkArea> work = 
   new ContextLocal<WorkArea>()
   {
      protected synchronized WorkArea initialValue()
      {
         return new WorkArea();
      }
   };
      
   /**
    * Prevent creation from outside of this class.
    */
   private WindowManager()
   {
   }

   /**
    * Compute the child hierarchy for the specified window.
    * 
    * @param    parentId
    *           The ID of the parent window.
    * @param    reverse
    *           When <true>code</true>, the window list is walked in reverse order.
    *           
    * @return   The list of child windows, in a depth-first search.
    */
   public static List<TitledWindow<?>> findChildren(int parentId, boolean reverse)
   {
      WorkArea wa = work.get();
      List<TitledWindow<?>> result = new ArrayList<>();
      
      // this lock is used for both WorkArea.windows and WorkArea.fixedOrderWindows fields
      synchronized (wa.windows)
      {
         // use a depth-first search; we can't use the WorkArea.windows, as this does not provide
         // a stable order (as this is used for the windows' z-order). 
         // instead, the resources can be walked using their creation order, as fortunately, the
         // WINDOW's FIRST-CHILD attribute depends on this, and not on the order on which the
         // PARENT attribute is set.

         List<TitledWindow<?>> wins = new ArrayList<>();
         wins.addAll(wa.fixedOrderWindows);
         if (reverse)
         {
            Collections.reverse(wins);
         }

         for (TitledWindow<?> tw : wins)
         {
            WidgetConfig wcfg = tw.config();
            
            if (wcfg instanceof BaseConfig && ((BaseConfig) wcfg).parentId == parentId)
            {
               result.addAll(findChildren(wcfg.id.asInt(), reverse));
               result.add(tw);
            }
         }
      }
      
      return result;
   }
   
   /**
    * Add top level window.
    * 
    * @param   window
    *          Window to add.
    */
   @SuppressWarnings("rawtypes")
   public static void addWindow(TitledWindow window)
   {
      WorkArea wa = work.get();
      
      synchronized (wa.windows)
      {
         if (!wa.windows.contains(window))
         {
            wa.windows.add(window);
            wa.fixedOrderWindows.add(window);
         }
      }
   }
   
   /**
    * Move the specified window to the bottom, behind any other windows.
    * 
    * @param    window
    *           The window to be moved.
    */
   @SuppressWarnings("rawtypes")
   public static void moveToBottom(TitledWindow window)
   {
      WorkArea wa = work.get();
      
      synchronized (wa.windows)
      {
         wa.windows.remove(window);
         
         wa.windows.add(0, window);
      }
   }
   
   /**
    * Move the specified window to the top, above any other windows.
    * 
    * @param    window
    *           The window to be moved.
    */
   @SuppressWarnings("rawtypes")
   public static void moveToTop(TitledWindow window)
   {
      WorkArea wa = work.get();
      
      synchronized (wa.windows)
      {
         wa.windows.remove(window);
         
         wa.windows.add(window);
      }
   }
   
   /**
    * Find the window with the given ID.
    * 
    * @param    windowId
    *           The window ID.
    */
   @SuppressWarnings({ "rawtypes", "unchecked" })
   public static <T extends TitledWindow> T findWindow(int windowId)
   {
      WorkArea wa = work.get();
      
      synchronized (wa.windows)
      {
         for (TitledWindow tw : wa.windows)
         {
            WidgetId id = tw.getId();
            
            if (id != null && id.asInt() == windowId)
            {
               return (T) tw;
            }
         }
      }
      
      return null;
   }

   /**
    * Delete all windows.
    */
   public static void clearWindowList()
   {
      WorkArea wa = work.get();
      
      @SuppressWarnings("rawtypes")
      List<TitledWindow> wins = new ArrayList<>();

      synchronized (wa.windows)
      {
         wins.addAll(wa.windows);
      }

      // notify the windows are being destroyed
      for (TitledWindow<?> w : wins)
      {
         w.destroy();
      }

      synchronized (wa.windows)
      {
         wa.windows.clear();
         wa.fixedOrderWindows.clear();
      }
   }

   /**
    * Find focused widget.
    * 
    * @return  focused widget or <code>null</code> if there is no such widget.
    */
   @SuppressWarnings("rawtypes")
   public static Widget findFocus()
   {
      Widget currentFocus;
      
      WorkArea wa = work.get();
      
      synchronized (wa.windows)
      {
         currentFocus = topWindow().currentFocus();

         while (currentFocus instanceof Container)
         {
            Widget childFocus = ((Container) currentFocus).currentFocus();

            if (childFocus == currentFocus || childFocus == null)
               break;

            currentFocus = childFocus;
         }

         if (currentFocus == null)
            currentFocus = topWindow();
      }
      return currentFocus;
   }

   /**
    * Return top window in Z-order.
    * 
    * @return  top window or <code>null</code> if no windows are present.
    */
   @SuppressWarnings("rawtypes")
   public static TitledWindow topWindow()
   {
      WorkArea wa = work.get();
      
      if (wa.windows.isEmpty())
         return null;

      int i = wa.windows.size() - 1;
      boolean isChui = ThinClient.getInstance().isChui();
      
      while (i >= 0)
      {
         TitledWindow w = wa.windows.get(i);
         
         // ChUI has a peculiarity here... it requires the top window (if is a message-box or 
         // something else) regardless of visibility state.
         if (isChui || w.isVisible())
         {
            return w;
         }
         i = i - 1;
      }
      
      // TODO: what is the correct behaviour?
      return null;
   }

   /**
    * Get list of top level windows.
    * 
    * @return  list of top level windows.
    */
   @SuppressWarnings("rawtypes")
   public static List<TitledWindow<?>> windowList()
   {
      return Collections.unmodifiableList(work.get().windows);
   }

   /**
    * Check if specified window is displayed, i.e. present in the list of 
    * windows.
    * 
    * @param   window
    *          Window to check.
    *          
    * @return  <code>true</code> if window is displayed.
    */
   @SuppressWarnings("rawtypes")
   public static boolean isDisplayed(Widget window)
   {
      WorkArea wa = work.get();
      
      synchronized (wa.windows)
      {
         return wa.windows.contains(window);
      }
   }

   /**
    * Remove specified window from window list.
    * 
    * @param   window
    *          Window to remove.
    */
   public static void remove(TitledWindow<?> window)
   {
      WorkArea wa = work.get();
      
      // this needs to be done out of lock
      window.destroy();

      synchronized (wa.windows)
      {
         wa.windows.remove(window);
         wa.fixedOrderWindows.remove(window);
      }
   }

   /**
    * Check if specified rectangle intersects at least partially with any of 
    * the windows located above specified one in the Z-order.
    * 
    * @param   bounds
    *          Source rectangle.
    * @param   ancestor
    *          Window to check.
    *          
    * @return  <code>true</code> if there is an intersection.
    */
   @SuppressWarnings("rawtypes")
   public static boolean checkRectangle(Rectangle bounds, Widget ancestor)
   {
      List<TitledWindow<?>> list = windowList();
      
      WorkArea wa = work.get();

      synchronized (list)
      {
         int i = wa.windows.indexOf(ancestor);

         if (i < 0)
            return false;
    
         if (!ThinClient.getInstance().isChui())
         {
            // in GUI, all windows use their own canvas.  In ChUI, the canvas/screen is shared, 
            // so intersection needs to be checked.
            return false;
         }
         
         i++;

         for (; i < list.size(); i++)
         {
            if (bounds.equals(list.get(i).bounds().intersection(bounds)))
               return true;
         }
      }

      return false;
   }
   
   /**
    * Get the list of all window IDs.
    * 
    * @return   See above.
    */
   public static int[] windowIds()
   {
      WorkArea wa = work.get();
      
      int[] ids = new int[wa.windows.size()];
      int i = 0;
      for (TitledWindow<?> w : wa.windows)
      {
         ids[i++] = w.getId().asInt();
      }
      
      return ids;
   }
   
   /**
    * Place the new window on the screen, in its default location.
    * <p>
    * The window placement algorithm replicates the algorithm used by WIN32 default window 
    * placement.  More details can be found in note 615 of Redmine task #2252.
    * 
    * @param    window
    *           The window to be placed.
    */
   public static void placeWindow(Window<?> window)
   {
      WorkArea wa = work.get();
      
      Window<?> defw = Window.getDefaultWindow();
      
      if (window == defw)
      {
         // default window is not moved
         return;
      }

      // advance the default position by aligning it this way:
      // vertically: just below the window's title bar
      // horizontally: just right of the window's icon
      
      // always use the default window's metrics - the window title bar always has the same
      // metrics as the default one
      
      // if the new X and Y position doesn't allow the window to be fully displayed, reset to 0
      NativeDimension dim = OutputManager.getDriver().getScreenSize();
      NativeDimension wdim = window.physicalDimension();
      
      if (wa.x + wdim.width > dim.width || wa.y + wdim.height > dim.height)
      {
         wa.x = 0;
         wa.y = 0;
      }
      
      // initial position is always zero, so use the actual values as deltas
      window.moveBy(wa.x, wa.y);

      wa.x += wa.locationDx;
      wa.y += wa.locationDy;
   }   

   /**
    * Set the initial dimension for this window.  Can be overwritten by the business logic.
    * 
    * @param    window
    *           The window.
    */
   public static void setInitialDimension(Window<?> window)
   {
      WorkArea wa = work.get();

      window.config().widthChars = wa.defaultColumns;
      window.config().heightChars = wa.defaultRows;
   }

   /**
    * Set the default window dimension.
    * 
    * @param    rows
    *           The number of rows (character units).
    * @param    cols
    *           The number of columns (character units).
    */
   public static void setDefaultDimension(int rows, int cols)
   {
      WorkArea wa = work.get();
      wa.defaultRows = rows;
      wa.defaultColumns = cols;
   }
   
   /**
    * Set the window placement deltas.
    * 
    * @param   dx
    *          The X-axis advancement.
    * @param   dy
    *          The Y-axis advancement.
    */
   public static void setPlacementDeltas(int dx, int dy)
   {
      WorkArea wa = work.get();
      wa.locationDx = dx;
      wa.locationDy = dy;
   }

   /** 
    * Stores global data relating to the state of the current context.
    */
   private static class WorkArea
   {  
      /** Storage for windows. */
      private List<TitledWindow<?>> windows = new ArrayList<>();

      /** Storage for windows, following their creation order. */
      private List<TitledWindow<?>> fixedOrderWindows = new ArrayList<>();

      /** The default number of rows for new windows. */
      private int defaultRows;
      
      /** The default number of columns for new windows. */
      private int defaultColumns;
      
      /** The default X location for new windows. */
      private int x = 0;
      
      /** The default Y location for new windows. */
      private int y = 0;

      /** The X delta when placing a new window. */
      private int locationDy;

      /** The Y delta when placing a new window. */
      private int locationDx;
   }
}
