/*
** Module   : MenuChuiImpl.java
** Abstract : CHUI client implementation of Menu.
**
** Copyright (c) 2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -----------------------------Description-----------------------------
** 001 VIG 20150116 Created initial version.
** 002 HC  20150323 Removed business logic from config classes.
** 003 VIG 20150225 Added event processing and drawing methods.
*/
package com.goldencode.p2j.ui.client.chui;

import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.border.*;
import com.goldencode.p2j.ui.client.chui.driver.ChuiOutputManager;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * <code>Sub-menu</code> CHUI client implementation class. 
 */
@SuppressWarnings("unchecked")
public class SubMenuChuiImpl
extends SubMenu<ChuiOutputManager>
{
   /** Border around the panel. */
   private Border<ChuiOutputManager> border = null;
   
   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public SubMenuChuiImpl(WidgetId id, SubMenuConfig cfg)
   {
      super(id, cfg);
      
      border = new LineBorder()
      {
         /**
          * Draw line border.
          * 
          * @param   widget
          *          Base widget. 
          * @param   size
          *          Border dimensions.
          */
         @Override
         public void drawBorder(Widget<ChuiOutputManager> widget, Dimension size)
         {
            widget.screen().box(bodyLocation(), 
                     size, 
                     widget.screen().borderColor(widget));
         }
      };
      
      //body = new BorderedPanel(new BorderLayout(), border);
   }
   
   /**
    * Drawing worker method.
    */
   public void drawPanel()
   {
      screen().clear(WidgetId._DEFAULT_WINDOW_ID, 
                     bodyLocation(), 
                     dimension(false, false), 
                     screen().getInteractiveOutput().widgetColor(this));

      if (border != null)
         border.drawBorder(this, dimension(false, false));
   }

   /**
    * Lay out container.
    */
   @Override
   public void doLayout()
   {
      super.doLayout();

      Widget<ChuiOutputManager>[] w = widgets();

      for (int i = 0; i < w.length; i++)
      {
         w[i].setLocation(new Point(1, i));

         if (w[i] instanceof SubMenu)
         {
            ((SubMenu<?>) w[i]).doLayout();
         }
      }
   }
   
   /**
    * Draw the peer container. The methods calls the draw() method for
    * ScrollPane instance for the peer container. In addition the required
    * clipping rectangle is calculated for ChildFrame containers because they
    * must be rendered inside the parent container.
    * <p>
    * Drawing for the redirected terminal differs. Labels/headers are drawn
    * only once per page, instead of the usual once per menu.
    */
   @Override
   public void draw()
   {
      Point childrenLoc = parent().childrenLocation(this);
      Point loc = location();
      drawTitle(new Point(loc.x + childrenLoc.x, loc.y + childrenLoc.y));

      if (config().displayed)
      {
         drawPanel();
         super.draw();
      }
   }
   
   /**
    * Hide/show widget.
    * <p>
    * If the menu is being shown and not yet attached to a window, attached it to the default
    * window.
    * 
    * @param   visible
    *          New visible state.
    */
   @Override
   public void setVisible(boolean visible)
   {
      // SUB-MENU is not displayed (dropped down) by default
      // setDisplayed(visible);
      
      for(Widget<ChuiOutputManager> w : widgets())
      {
         w.setVisible(true);
      }

      super.setVisible(visible);
   }

   /**
    * Replace original version to take into account P2J specific behavior.
    * 
    * @see  com.goldencode.p2j.ui.client.widget.Widget#dimension()
    */
   @Override
   public Dimension dimension()
   {
      return dimension(true, true);
   }

   /**
    * Get children location for one child.
    * 
    * @param   child
    *          Child component.
    *          
    * @return  Point - parents location
    */
   @Override
   public Point childrenLocation(Widget<ChuiOutputManager> child)
   {
      Point bodyLoc = bodyLocation();
      return new Point(bodyLoc.x + 1, bodyLoc.y + 1); // border
   }

   /**
    * Get widget width.
    * 
    * @return  widget width.
    */
   @Override
   public double width()
   {
      double w = titleWidth();
      
      if (config().displayed)
      {
         w = Math.max(w, SubMenuChuiImpl.getMaxItemWidth(this, true) + 2); // borders
      }
      
      return w;
   }

   /**
    * Get sub-menu height.
    * 
    * @return  sub-menu height.
    */
   @Override
   public double height()
   {
      double h = 1; // title height
      
      if (config().displayed)
      {
         h = getMaxItemHeight(this, true);
      }
      
      return h;
   }

   /**
    * Check if widget needs in cursor to be on when having the input focus.  By default this
    * method always returns <code>true</code> so it needs to be overwritten if other return value
    * is required for particular widget implementation class.
    * 
    * @return  <code>true</code> as the default value.
    */
   @Override
   public boolean forceCursorOnFocus()
   {
      return false;
   }

   /**
    * Check if the current widget can manage non-global system actions on its own, and not via the
    * enclosing event loop.
    * 
    * @return   <code>true</code> as the widget has its own event management.
    */
   @Override
   public boolean manageSystemActions()
   {
      return true;
   }

   /**
    * Calculates a max height of specified Menu or SubMenu.
    * Analyzes nested dropped-down sub-menus.
    * 
    * @param   menu
    *          Menu or SubMenu.
    *          
    * @param   includeNested
    *          <code>true</code> process nested sub-menus.
    * 
    * @return  a height of menu.
    */
   static double getMaxItemHeight(AbstractContainer<?> menu, boolean includeNested)
   {
      Widget<?>[] widgets = menu.widgets();
      double height = widgets.length + 2; // borders
      
      if (includeNested)
      {
         for (int i = 0; i < widgets.length; i++)
         {
            if (widgets[i] instanceof SubMenu)
            {
               SubMenu<?> sm = (SubMenu<?>) widgets[i];
               if (sm.config().displayed)
               {
                  height = Math.max(height, sm.height() + i + 1); // this.y + title height
               }
            }
         }
      }
      return height;
   }
   
   /**
    * Calculates max menu/sub-menu item width
    * 
    * @param    menu
    *           MENU or SUB-MENU
    * @param    includeNested
    *           <code>true</code> means, that we should process nested sub-menus, 
    *           <code>false</code> otherwise.          
    *           
    * @return   max width.
    */
   static double getMaxItemWidth(Container<?> menu, boolean includeNested)
   {
      double width = 0; 
      for(Widget<?> w : menu.widgets())
      {
         AbstractWidget<?> aw = (AbstractWidget<?>) w;
         
         // ignore SKIP or RULE
         if (aw instanceof MenuItem && !((MenuItem<?>)aw).isNormal())
         {
            continue;
         }
         
         double itemWidth;
         if (aw instanceof SubMenu && !includeNested)
         {
            itemWidth = ((SubMenu<?>)aw).titleWidth();
         }
         else
         {
            itemWidth = aw.width();
         }
         
         if (width < itemWidth)
         {
            width = itemWidth; 
         }
      }
      return width;
   }

   /**
    * Prepare widgets before drawing
    * 
    * @param   widgets
    *          Container children.
    *          
    * @return  widgets, prepared for drawing.
    */
   @Override
   protected List<Widget<ChuiOutputManager>> prepareWidgets(List<Widget<ChuiOutputManager>> widgets)
   {
      List<Widget<ChuiOutputManager>> result = new ArrayList<>();
      Widget<ChuiOutputManager> displayedSubmenu = null;
      
      for (Widget<ChuiOutputManager> w : widgets)
      {
         if (w instanceof SubMenu && ((SubMenu<?>)w).config().displayed)
         {
            displayedSubmenu = w;
         }
         else
         {
            result.add(w);
         }
      }
      
      if (displayedSubmenu != null)
      {
         result.add(displayedSubmenu);
      }
      
      return Collections.unmodifiableList(result);
   }
   
   /**
    * Get location of SUB-MENU body, relative to container.
    * 
    * @return  body location.
    */
   @Override
   protected Point bodyLocation()
   {
      Point loc = parent().childrenLocation(this).translate(location());
      return new Point(loc.x, loc.y + 1); // under the title
   }

   /**
    * Trick, that lets to correct parent location in cases, when it needed,
    * for example for menus.
    * 
    * @param    parent
    *           Parent component.
    *           
    * @return   corrected parent location.
    */
   @Override
   protected Point prepParentLocation(Widget<?> parent)
   {
      Point loc = parent.location();
      if (parent instanceof SubMenu)
      {
         loc.x += 1; // border
         loc.y += 2; // top border and title
      }
      return loc;
   }
   
   /**
    * Draws menu-item title, drawing mnemonic character as underlined.
    * 
    * @param   p
    *          Title point.  
    */
   private void drawTitle(Point p)
   {
      Color dColor = hasFocus() ? Color.MESSAGES : Color.NORMAL; // highlighted() ?
      Color mnemonicColor = Color.INPUT;
      
      String title = this.config().title;
      int ampIdx = title.indexOf('&');
      boolean amp = ampIdx != -1;
      // if & is not specified use the first letter as mnemonic
      if (!amp) ampIdx = 0;
      
      Point tp = new Point(p);
      
      String leftPart = title.substring(0, ampIdx);
      screen().at(tp).append(leftPart, dColor);

      tp.x += ampIdx;
      // "eat" the ampersand
      if (amp) ampIdx++;
      String mnemonic = title.substring(ampIdx, ampIdx + 1);
      screen().at(tp).append(mnemonic, mnemonicColor);

      tp.x += 1;
      String rightPart = title.substring(ampIdx + 1);
      screen().at(tp).append(rightPart, dColor);

      tp.x += rightPart.length();
      
      if (!Menu.isMenubarElement(this))
      {
         screen().at(tp).append(" ->", dColor);
      }
   }

   /**
    * Replace original version to take into account P2J specific behavior.
    * 
    * @param   includeNested
    *          <code>true</code> calculates dimension processing nested sub-menus.
    * @param   includeTitle
    *          <code>true</code> add title height (1).
    * 
    * @see  com.goldencode.p2j.ui.client.widget.Widget#dimension()
    */
   private Dimension dimension(boolean includeNested, boolean includeTitle)
   {
      if (!config().displayed)
      {
         return new Dimension(width(), 1); 
      }
      else
      {
         double width = getMaxItemWidth(this, includeNested);

         int height = widgets().length + 2; // borders
         
         if (includeTitle)
            height++;
         
         return new Dimension(width + 4, height); // borders and title
      }
   }
   
}
